@@ -75,6 +75,8 @@ use trust_dns_resolver::config::{NameServerConfig, Protocol, ResolverConfig};
 use crate::acl::AccessControl;
 #[cfg(feature = "local-dns")]
 use crate::local::dns::NameServerAddr;
+#[cfg(feature = "local")]
+use crate::local::socks::config::Socks5AuthConfig;
 
 #[derive(Serialize, Deserialize, Debug)]
 #[serde(untagged)]
@@ -246,6 +248,11 @@ struct SSLocalExtConfig {
     #[cfg(feature = "local-tun")]
     #[serde(skip_serializing_if = "Option::is_none")]
     tun_interface_address: Option<String>,
+
+    /// SOCKS5
+    #[cfg(feature = "local")]
+    #[serde(skip_serializing_if = "Option::is_none")]
+    socks5_auth_config_path: Option<String>,
 }
 
 #[derive(Serialize, Deserialize, Debug)]
@@ -823,6 +830,10 @@ pub struct LocalConfig {
 
     /// Set `IPV6_V6ONLY` for listener socket
     pub ipv6_only: bool,
+
+    /// SOCKS5 Authentication configuration
+    #[cfg(feature = "local")]
+    pub socks5_auth: Socks5AuthConfig,
 }
 
 impl LocalConfig {
@@ -859,6 +870,9 @@ impl LocalConfig {
             tun_device_fd_from_path: None,
 
             ipv6_only: false,
+
+            #[cfg(feature = "local")]
+            socks5_auth: Socks5AuthConfig::default(),
         }
     }
 
@@ -1420,6 +1434,11 @@ impl Config {
                             local_config.tun_interface_name = Some(tun_interface_name);
                         }
 
+                        #[cfg(feature = "local")]
+                        if let Some(socks5_auth_config_path) = local.socks5_auth_config_path {
+                            local_config.socks5_auth = Socks5AuthConfig::load_from_file(&socks5_auth_config_path)?;
+                        }
+
                         nconfig.local.push(local_config);
                     }
                 }
@@ -2118,6 +2137,9 @@ impl fmt::Display for Config {
                         tun_interface_name: local.tun_interface_name.clone(),
                         #[cfg(feature = "local-tun")]
                         tun_interface_address: local.tun_interface_address.as_ref().map(ToString::to_string),
+
+                        #[cfg(feature = "local")]
+                        socks5_auth_config_path: None,
                     };
                     jlocals.push(jlocal);
                 }
