@@ -1,10 +1,14 @@
 //! UDP Tunnel server
 
-use std::{io, net::SocketAddr, sync::Arc, time::Duration};
+use std::{
+    io::{self, ErrorKind},
+    net::SocketAddr,
+    sync::Arc,
+    time::Duration,
+};
 
 use bytes::Bytes;
 use futures::future;
-use io::ErrorKind;
 use log::{debug, error, info, trace, warn};
 use lru_time_cache::LruCache;
 use shadowsocks::{
@@ -16,73 +20,39 @@ use shadowsocks::{
     },
     ServerAddr,
 };
-use tokio::{
-    net::UdpSocket,
-    sync::{mpsc, Mutex},
-    task::JoinHandle,
-    time,
-};
+use tokio::{net::UdpSocket, sync::mpsc, task::JoinHandle, time};
 
 use crate::{
     local::{context::ServiceContext, loadbalancing::PingBalancer},
-    net::MonProxySocket,
+    net::{MonProxySocket, UDP_ASSOCIATION_KEEP_ALIVE_CHANNEL_SIZE, UDP_ASSOCIATION_SEND_CHANNEL_SIZE},
 };
 
 type AssociationMap = LruCache<SocketAddr, UdpAssociation>;
-type SharedAssociationMap = Arc<Mutex<AssociationMap>>;
 
 pub struct UdpTunnel {
     context: Arc<ServiceContext>,
-    assoc_map: SharedAssociationMap,
-    cleanup_abortable: JoinHandle<()>,
-    keepalive_abortable: JoinHandle<()>,
+    assoc_map: AssociationMap,
     keepalive_tx: mpsc::Sender<SocketAddr>,
-}
-
-impl Drop for UdpTunnel {
-    fn drop(&mut self) {
-        self.cleanup_abortable.abort();
-        self.keepalive_abortable.abort();
-    }
+    keepalive_rx: mpsc::Receiver<SocketAddr>,
+    time_to_live: Duration,
 }
 
 impl UdpTunnel {
     pub fn new(context: Arc<ServiceContext>, time_to_live: Option<Duration>, capacity: Option<usize>) -> UdpTunnel {
         let time_to_live = time_to_live.unwrap_or(crate::DEFAULT_UDP_EXPIRY_DURATION);
-        let assoc_map = Arc::new(Mutex::new(match capacity {
+        let assoc_map = match capacity {
             Some(capacity) => LruCache::with_expiry_duration_and_capacity(time_to_live, capacity),
             None => LruCache::with_expiry_duration(time_to_live),
-        }));
-
-        let cleanup_abortable = {
-            let assoc_map = assoc_map.clone();
-            tokio::spawn(async move {
-                loop {
-                    time::sleep(time_to_live).await;
-
-                    // cleanup expired associations. iter() will remove expired elements
-                    let _ = assoc_map.lock().await.iter();
-                }
-            })
         };
 
-        let (keepalive_tx, mut keepalive_rx) = mpsc::channel(64);
-
-        let keepalive_abortable = {
-            let assoc_map = assoc_map.clone();
-            tokio::spawn(async move {
-                while let Some(peer_addr) = keepalive_rx.recv().await {
-                    assoc_map.lock().await.get(&peer_addr);
-                }
-            })
-        };
+        let (keepalive_tx, keepalive_rx) = mpsc::channel(UDP_ASSOCIATION_KEEP_ALIVE_CHANNEL_SIZE);
 
         UdpTunnel {
             context,
             assoc_map,
-            cleanup_abortable,
-            keepalive_abortable,
             keepalive_tx,
+            keepalive_rx,
+            time_to_live,
         }
     }
 
@@ -110,28 +80,55 @@ impl UdpTunnel {
         let listener = Arc::new(socket);
 
         let mut buffer = [0u8; MAXIMUM_UDP_PAYLOAD_SIZE];
-        loop {
-            let (n, peer_addr) = match listener.recv_from(&mut buffer).await {
-                Ok(s) => s,
-                Err(err) => {
-                    error!("udp server recv_from failed with error: {}", err);
-                    time::sleep(Duration::from_secs(1)).await;
-                    continue;
-                }
-            };
+        let mut cleanup_timer = time::interval(self.time_to_live);
 
-            let data = &buffer[..n];
-            if let Err(err) = self
-                .send_packet(&listener, peer_addr, &balancer, forward_addr, data)
-                .await
-            {
-                error!(
-                    "udp packet relay {} -> {} with {} bytes failed, error: {}",
-                    peer_addr,
-                    forward_addr,
-                    data.len(),
-                    err
-                );
+        loop {
+            tokio::select! {
+                _ = cleanup_timer.tick() => {
+                    // cleanup expired associations. iter() will remove expired elements
+                    let _ = self.assoc_map.iter();
+                }
+
+                peer_addr_opt = self.keepalive_rx.recv() => {
+                    let peer_addr = peer_addr_opt.expect("keep-alive channel closed unexpectly");
+                    self.assoc_map.get(&peer_addr);
+                }
+
+                recv_result = listener.recv_from(&mut buffer) => {
+                    let (n, peer_addr) = match recv_result {
+                        Ok(s) => s,
+                        Err(err) => {
+                            error!("udp server recv_from failed with error: {}", err);
+                            time::sleep(Duration::from_secs(1)).await;
+                            continue;
+                        }
+                    };
+
+                    if n == 0 {
+                        // For windows, it will generate a ICMP Port Unreachable Message
+                        // https://docs.microsoft.com/en-us/windows/win32/api/winsock2/nf-winsock2-recvfrom
+                        // Which will result in recv_from return 0.
+                        //
+                        // It cannot be solved here, because `WSAGetLastError` is already set.
+                        //
+                        // See `relay::udprelay::utils::create_socket` for more detail.
+                        continue;
+                    }
+
+                    let data = &buffer[..n];
+                    if let Err(err) = self
+                        .send_packet(&listener, peer_addr, &balancer, forward_addr, data)
+                        .await
+                    {
+                        error!(
+                            "udp packet relay {} -> {} with {} bytes failed, error: {}",
+                            peer_addr,
+                            forward_addr,
+                            data.len(),
+                            err
+                        );
+                    }
+                }
             }
         }
     }
@@ -144,9 +141,7 @@ impl UdpTunnel {
         forward_addr: &Address,
         data: &[u8],
     ) -> io::Result<()> {
-        let mut assoc_map = self.assoc_map.lock().await;
-
-        if let Some(assoc) = assoc_map.get(&peer_addr) {
+        if let Some(assoc) = self.assoc_map.get(&peer_addr) {
             return assoc.try_send(Bytes::copy_from_slice(data));
         }
 
@@ -162,7 +157,7 @@ impl UdpTunnel {
         debug!("created udp association for {}", peer_addr);
 
         assoc.try_send(Bytes::copy_from_slice(data))?;
-        assoc_map.insert(peer_addr, assoc);
+        self.assoc_map.insert(peer_addr, assoc);
 
         Ok(())
     }
@@ -208,6 +203,7 @@ struct UdpAssociationContext {
     forward_addr: Address,
     proxied_socket: Option<MonProxySocket>,
     keepalive_tx: mpsc::Sender<SocketAddr>,
+    keepalive_flag: bool,
     balancer: PingBalancer,
     inbound: Arc<UdpSocket>,
 }
@@ -227,10 +223,10 @@ impl UdpAssociationContext {
         keepalive_tx: mpsc::Sender<SocketAddr>,
         balancer: PingBalancer,
     ) -> (JoinHandle<()>, mpsc::Sender<Bytes>) {
-        // Pending packets 128 for each association should be good enough for a server.
+        // Pending packets UDP_ASSOCIATION_SEND_CHANNEL_SIZE for each association should be good enough for a server.
         // If there are plenty of packets stuck in the channel, dropping excessive packets is a good way to protect the server from
         // being OOM.
-        let (sender, receiver) = mpsc::channel(128);
+        let (sender, receiver) = mpsc::channel(UDP_ASSOCIATION_SEND_CHANNEL_SIZE);
 
         let mut assoc = UdpAssociationContext {
             context,
@@ -238,6 +234,7 @@ impl UdpAssociationContext {
             forward_addr,
             proxied_socket: None,
             keepalive_tx,
+            keepalive_flag: false,
             balancer,
             inbound,
         };
@@ -248,6 +245,7 @@ impl UdpAssociationContext {
 
     async fn dispatch_packet(&mut self, mut receiver: mpsc::Receiver<Bytes>) {
         let mut proxied_buffer = Vec::new();
+        let mut keepalive_interval = time::interval(Duration::from_secs(1));
 
         loop {
             tokio::select! {
@@ -276,6 +274,16 @@ impl UdpAssociationContext {
 
                     self.send_received_respond_packet(&addr, &proxied_buffer[..n]).await;
                 }
+
+                _ = keepalive_interval.tick() => {
+                    if self.keepalive_flag {
+                        if let Err(..) = self.keepalive_tx.try_send(self.peer_addr) {
+                            debug!("udp relay {} keep-alive failed, channel full or closed", self.peer_addr);
+                        } else {
+                            self.keepalive_flag = false;
+                        }
+                    }
+                }
             }
         }
 
@@ -353,18 +361,19 @@ impl UdpAssociationContext {
     }
 
     async fn send_received_respond_packet(&mut self, addr: &Address, data: &[u8]) {
-        trace!("udp relay {} <- {} received {} bytes", self.peer_addr, addr, data.len(),);
+        trace!("udp relay {} <- {} received {} bytes", self.peer_addr, addr, data.len());
+
         // Keep association alive in map
-        let _ = self
-            .keepalive_tx
-            .send_timeout(self.peer_addr, Duration::from_secs(1))
-            .await;
+        self.keepalive_flag = true;
 
         // Send back to client
         if let Err(err) = self.inbound.send_to(data, self.peer_addr).await {
             warn!(
-                "udp failed to send back to client {}, from target {}, error: {}",
-                self.peer_addr, addr, err
+                "udp failed to send back {} bytes to client {}, from target {}, error: {}",
+                data.len(),
+                self.peer_addr,
+                addr,
+                err
             );
         } else {
             trace!("udp relay {} <- {} with {} bytes", self.peer_addr, addr, data.len());
